from __future__ import absolute_import, division, print_function

import numpy as np
import torch
from torch.autograd import Variable, grad
from detox.attack import Attack
from detox.util import clip_eta


class FeatureAdversaries(Attack):
    """
    Produces adversarial image based on internal layers of network
    representations.
    tags - white box, targeted, iterative, image specific
    Reference:
        Adversarial Manipulation of Deep Representations,
        Sabour et al., https://arxiv.org/abs/1511.05122
    """
    def __init__(self, model, layer, eps=0.3, eps_iter=0.5, ord=np.inf, n_iter=10,
                 clip_min=None, clip_max=None, **kwargs):
        """
        Args:
            model (nn.Module): must return the output and layer output
            layer (int): index of the layer to attack
        """
        super(FeatureAdversaries, self).__init__()
        if not isinstance(model, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.model = model
        self.layer = layer
        self.eps = eps
        self.eps_iter = eps_iter
        self.ord = ord
        self.n_iter = n_iter
        self.clip_min = clip_min
        self.clip_max = clip_max

    def generate(self, x, g):
        """
        x: source
        g: guide
        """
        eta = torch.randn(x.size())
        eta = clip_eta(eta, self.ord, self.eps)

        _, g_feat = self.model(g)

        # run iterations
        for i in range(self.n_iter):
            eta = self._step(x, eta, g_feat)

        adv_x = x + eta
        # clip by the range [clip_min, clip_max]
        if (self.clip_min is not None) and (self.clip_max is not None):
            adv_x = torch.clamp(adv_x, self.clip_min, self.clip_max)
        return adv_x

    def _step(self, x, eta, g_feat):
        """
        Args:
            x (Variable): Input variable
            eta (Variable): Perturbation tensor with the same shape of x
            g_feat (Variable): internal tensor for guide
        """
        adv_x = Variable(x.data + eta, requires_grad=True)
        _, a_feat = self.model(adv_x)

        # negative for targeted attack
        loss = -torch.sum((a_feat - g_feat)**2))

        # gradient wrt adv_x & multiply with epsilon
        gradient = grad(loss, adv_x)[0]
        scaled_grad = self.eps_iter * torch.sign(gradient)

        adv_x = adv_x + scaled_grad
        if (self.clip_min is not None) and (self.clip_max is not None):
            adv_x = torch.clamp(adv_x, self.clip_min, self.clip_max)

        # repackage
        adv_x = Variable(adv_x.data)
        eta = adv_x - x
        eta = clip_eta(eta, self.ord, self.eps)

        return eta



from __future__ import absolute_import, division, print_function

from detox.attack.attack import Attack
from detox.attack.fgm import FGSM, FGM
from detox.attack.basic_iterative import BasicIterative
from detox.attack.momentum_iterative import MomentumIterative
from detox.attack.jsma import JSMA
from detox.attack.one_pixel import OnePixel
from detox.attack.deep_fool import DeepFool
from detox.attack.atn import ATN
# from detox.attack.carlini import CarliniL2, CarliniLi
# from detox.attack.feature_adversaries import FeatureAdversaries

from detox.attack import functional

__all__ = [
    # Bases
    'Attack',
    # Attack Classes
    'FGSM',
    'FGM',
    'JSMA',
    'BasicIterative',
    'MomentumIterative',
    'OnePixel',
    'DeepFool',
    'ATN',
    # Functionals
    'functional',
]



from __future__ import absolute_import, division, print_function

import contextlib
import re

import torch


def parse_torch_version():
    """
    Parses `torch.__version__` into a semver-ish version tuple.
    This is needed to handle subpatch `_n` parts outside of the semver spec.

    :returns: a tuple `(major, minor, patch, extra_stuff)`
    """
    match = re.match(r"(\d\.\d\.\d)(.*)", torch.__version__)
    major, minor, patch = map(int, match.group(1).split("."))
    extra_stuff = match.group(2)
    return major, minor, patch, extra_stuff


# Polyfill to bridge the change of .volatile between PyTorch 0.3 and 0.4.
try:
    # These work in PyTorch 0.4 prerelease.
    torch_no_grad = torch.no_grad

    def is_volatile(variable):
        return False

except AttributeError:
    # These work in PyTorch 0.3 and earlier.

    @contextlib.contextmanager
    def torch_no_grad():
        yield

    def is_volatile(variable):
        return variable.volatile



from __future__ import absolute_import, division, print_function

import torch


# Evaluation Functions

def topk_miss(probs, target, k):
    r"""Computes number of misses that are not included in top k predictions."""
    return probs.size(-1) - topk_correct(probs, target, k)


def topk_correct(probs, target, k):
    r"""Computes number of correct answers in top k."""
    _, pred = torch.topk(probs, k)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))
    correct_k = correct.view(-1).sum(0, keepdim=True)
    return correct_k.cpu()[0]

def topk_accuracy(probs, target, k):
    r"""Computes accuracy of the batch."""
    batch_size = target.size(0)
    correct_k = topk_correct(probs, target, k)
    return correct_k * (100.0 / batch_size)


class Metric(object):
    def __init__(self):
        pass

    def update(self):
        pass

    def reset(self):
        pass

    def accuracy(self):
        pass



from __future__ import absolute_import, division, print_function

import torch
import torch.nn as nn
from torch.nn import functional as F
from torch.autograd import Variable
from torch.optim import Adam
from ignite.trainer import Trainer

from detox.attack import Attack
from detox.nn import ae


class TextNAE(Attack):
    pass


class NAE(Attack):
    """
    Generates natural adversarial examples by searching in semantic space of
    dense and continuous data representation, which is learned by GAN.
    Reference:
        Generating Natural Adversarial Examples,
        Zhao et al., https://arxiv.org/abs/1710.11342
    """
    def __init__(self, model, G=None, D=None, inverter=None,
                 alpha=1.5, beta=0.1):
        super(NAE, self).__init__()

        if not isinstance(G, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.G = G
        if not isinstance(D, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.D = D

        if inverter is None:
            raise ValueError('inverter is not defined')
        elif generator is None:
            raise ValueError('generator is not defined')
        else:
            if not isinstance(inverter, nn.Module):
                raise ValueError('inverter must be an instance of nn.Module')
            if not isinstance(generator, nn.Module):
                raise ValueError('generator must be an instance of nn.Module')
        self.generator = generator
        self.inverter = inverter

        self.alpha = alpha
        self.beta = beta

    def grad_penalty(self, D, real, gen, context, lambd):
        alpha = torch.rand(real.size()).cuda()
        x_hat = alpha * real + ((1 - alpha) * gen).cuda()
        x_hat = Variable(x_hat, requires_grad=True)
        context = Variable(context)
        d_hat = D(x_hat, context)
        ones = torch.ones(d_hat.size()).cuda()
        gradients = grad(outputs=d_hat, inputs=x_hat,
                         grad_outputs=ones, create_graph=True,
                         retain_graph=True, only_inputs=True)[0]
        penalty = ((gradients.norm(2, dim=1) - 1) ** 2).mean() * lamb
        return penalty

    def criterion_g(self, D, G, src, trg, curriculum=None):
        src_len = min(curriculum, len(src)-1) + 1
        trg_len = min(curriculum, len(src)-1) + 1
        gen_trg, context = G(src[:src_len], trg[:trg_len])
        loss_g = D(gen_trg, context)
        loss = -loss_g.mean()
        return loss

    def criterion_d(self, D, G, src, trg, lambd, curriculum=None):
        src_len = min(curriculum, len(src)-1) + 1
        trg_len = min(curriculum, len(src)-1) + 1
        # with gen
        gen_trg, context = G(src[:src_len], trg[:trg_len])
        d_gen = D(gen_trg, context)
        # with real
        trg = to_onehot(trg, D.vocab_size).type(torch.FloatTensor)[1:trg_len]
        trg = Variable(trg.cuda())
        d_real = D(trg, context)
        # calculate gradient panalty
        penalty = grad_penalty(D, trg.data, gen_trg.data, context.data, lamb)
        loss = d_gen.mean() - d_real.mean() + penalty
        return loss

    def trainer(self, optimizer, target=None, cuda=False):

        def _prepare_batch(batch, volatile=False):
            x, y = batch
            if cuda:
                x, y = x.cuda(), y.cuda()
            return (Variable(x, volatile=volatile),
                    Variable(y, volatile=volatile))

        def _update(batch):
            self.adv_model.train()
            self.model.eval()
            optimizer.zero_grad()

            x, y = _prepare_batch(batch)
            adv_x = self.adv_model(x)
            logits = self.model(x)
            probs = F.softmax(logits, dim=1)
            probs = Variable(probs.data)

            reranked = self.rerank(probs, target)
            reranked = Variable(reranked.data)

            loss = self.criterion(x, adv_x, probs, reranked)
            loss.backward()
            optimizer.step()
            return loss.data[0]

        trainer = Trainer(_update)
        return trainer

    def generate(self, x):
        adv_x = self.adv_model(x)
        return adv_x



from __future__ import absolute_import, division, print_function

import torch
import torch.nn as nn
from torch.nn import functional as F
from torch.autograd import Variable
from torch.optim import Adam
from ignite.trainer import Trainer

from detox.attack import Attack
from detox.nn import autoencoder


class ATN(Attack):
    """
    Adversarial Transformation Networks
    Reference:
        Adversarial Transformation Networks:
        Learning to Generate Adversarial Examples,
        Baluja et al., https://arxiv.org/abs/1703.09387
    """
    def __init__(self, model, adv_model=None, alpha=1.5, beta=0.1):
        super(ATN, self).__init__(model)
        if not isinstance(model, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.model = model

        if adv_model is None:
            adv_model = autoencoder(pretrained=True)
        else:
            if not isinstance(adv_model, nn.Module):
                raise ValueError('Model must be an instance of nn.Module')
        self.adv_model = adv_model
        self.alpha = alpha
        self.beta = beta

    def criterion(self, x, adv_x, pred, rerank):
        Lx = torch.sum(
            torch.sqrt(torch.sum((adv_x - x)**2 + 1e-12, 1))
        )
        Ly = torch.sum(
            torch.sqrt(torch.sum((pred - rerank)**2 + 1e-12, 1))
        )
        return self.beta * Lx + Ly

    def rerank(self, probs, target):
        """ Reranking function that modifies y such that y_k < y_t.
        we use reranking function that maintains the rank order of all but the
        targeted class in order to minimize distortions"""
        max_val, _ = torch.max(probs, 1)
        probs[:, target] = max_val
        reranked = F.normalize(probs)
        return reranked

    def trainer(self, target, optimizer, cuda=False):

        def _prepare_batch(batch, volatile=False):
            x, y = batch
            if cuda:
                x, y = x.cuda(), y.cuda()
            return (Variable(x, volatile=volatile),
                    Variable(y, volatile=volatile))

        def _update(batch):
            self.adv_model.train()
            self.model.eval()
            optimizer.zero_grad()

            x, y = _prepare_batch(batch)
            adv_x = self.adv_model(x)
            logits = self.model(x)
            probs = F.softmax(logits, dim=1)
            probs = Variable(probs.data)

            reranked = self.rerank(probs, target)
            reranked = Variable(reranked.data)

            loss = self.criterion(x, adv_x, probs, reranked)
            loss.backward()
            optimizer.step()
            return loss.data[0]

        trainer = Trainer(_update)
        return trainer

    def generate(self, x):
        adv_x = self.adv_model(x)
        return adv_x, None



from __future__ import absolute_import, division, print_function

import torch
import numpy as np
from torch.autograd import grad, Variable
from torch.nn import functional as F
from detox.util import l2_batch_norm, kl_div
from torch.autograd.gradcheck import zero_gradients


def jacobian(x, target):
    """ Compute the Jacobian value corresponding to each pair of x and target.
    It helps us to estimate how changes in x affect each of the class
    probability.
    """
    if not x.requires_grad:
        raise ValueError('x must set require_grad to True')

    n_classes = target.size(1)
    jacobian_val = torch.zeros(n_classes, *x.size())
    grads = torch.zeros(*target.size())
    if x.is_cuda:
        jacobian_val = jacobian_val.cuda()
        grads = grads.cuda()

    # Compute gradients for all classes
    for i in range(n_classes):
        zero_gradients(x)
        grads.zero_()
        grads[:, i] = 1
        target.backward(grads, retain_variables=True)
        jacobian_val[i] = x.grad.data
    return jacobian_val.transpose(0, 1)


def saliency_map(jacobian, search_domain, target_index, increase=True):
    """ Computes Saliency Map, which determines the best pixels for our
    adversarial goal. It takes account of not just target class, but all other
    classes.
    Args:
        jacobian:
    """
    alpha = jacobian[target_index].squeeze()
    beta = torch.sum(jacobian, 0).squeeze() - alpha

    # Create a mask to keep features that match saliency map conditions
    if increase:
        m1, m2 = torch.ge(alpha, 0.0), torch.le(beta, 0.0)
    else:
        m1, m2 = torch.le(alpha, 0.0), torch.ge(beta, 0.0)
    mask = torch.mul(torch.mul(m1, m2), search_domain)

    if increase:
        scores = torch.mul(torch.mul(alpha, torch.abs(beta)), mask)
    else:
        scores = torch.mul(torch.mul(torch.abs(alpha), beta), mask)

    max_val, max_idx = torch.max(scores, dim=0)
    return max_val, max_idx


def fgsm(x, pred, eps=0.3, clip_min=None, clip_max=None):
    r""" Generates adversarial examples based on Fast Gradient Sign Method (FGSM)
    Args:
        x (Variable): Input variable.
        pred (Variable): Prediction of the model. (output of the softmax)
        y (Variable): (optional) One hot encoded target label.
        eps (float): Attack step size.
        ord (int): (optional) Order of the norm (np.inf, 1 or 2).
        clip_min (float): Minimum component value.
        clip_max (float): Maximum component value.
        targeted (bool): true or false
    Returns:
        adv_x (Variable):
            Generated adversarial example.
    Reference:
        Explaining and Harnessing Adversarial Examples,
        Goodfellow et al., CoRR2014, https://arxiv.org/abs/1412.6572
    """
    return fgm(x, pred, y=None, eps=eps, ord=np.inf,
               clip_min=clip_min, clip_max=clip_max)


def fgm(x, pred, y=None, eps=0.3, ord=np.inf,
        clip_min=None, clip_max=None, targeted=False):
    r""" Generates adversarial examples based on Fast Gradient Method (FGM)
    Args:
        x (Variable): Input variable.
        pred (Variable): Prediction of the model. (output of the softmax)
        y (Variable): (optional) One hot encoded target label.
        eps (float): Attack step size.
        ord (int): (optional) Order of the norm (np.inf, 1 or 2).
        clip_min (float): Minimum component value.
        clip_max (float): Maximum component value.
        targeted (bool): true or false
    Returns:
        adv_x (Variable):
            Generated adversarial example.
    Reference:
        Explaining and Harnessing Adversarial Examples,
        Goodfellow et al., CoRR2014, https://arxiv.org/abs/1412.6572
    """
    if y is None:
        # use model prediction as the label
        _, y = torch.max(pred, 1)
    # compute loss
    loss = F.cross_entropy(pred, y, size_average=False)
    if targeted:
        loss = -loss
    # get gradients wrt x
    gradient = grad(loss, x)[0]

    if ord == np.inf:
        # L-inf (sign)
        norm_grad = torch.sign(gradient)
    else:
        # Lp normalization
        norm_grad = F.normalize(gradient, ord)
    # add perturbation
    scaled_grad = eps * norm_grad
    adv_x = x + scaled_grad
    # clip by the range [clip_min, clip_max]
    if (clip_min is not None) and (clip_max is not None):
        adv_x = torch.clamp(adv_x, clip_min, clip_max)
    return adv_x


def lds(model, x, pred, eps=8., iterations=1, xi=1e-6,
        apply_softmax=True, batch_norm=True):
    r""" Computes local distributional smoothness (LDS) for Virtual Adversarial
    Training (VAT) with extension of l2 batch norm
    Args:
        model (nn.Module): The model that returns unnormalized logits.
        x (Variable): Input variable. (output of the softmax)
        logits (Variable): Output of the model (input to the softmax layer)
        eps (float): Attack step size.
        xi (float): the finite difference parameter
    Returns:
        adv_x (Variable):
            Generated adversarial example.
        lds (Variable):
            The LDS value.
    Reference:
        Distributional Smoothing With Virtual Adversarial Training
        https://arxiv.org/abs/1507.00677
    """
    y1 = Variable(pred.data)
    adv_x = vatm(model, x, pred, eps, iterations, xi,
                 apply_softmax, batch_norm)
    y2 = model(adv_x)
    if apply_softmax:
        y2 = F.softmax(y2, dim=1)
    return -kl_div(y1, y2)


def vatm(model, x, pred, eps=8., iterations=1, xi=1e-6,
         apply_softmax=True, batch_norm=True):
    r""" Generates adversarial sample based on Virtual Adversarial Training (VAT)
    Method with extension of l2 batch norm
    Args:
        model (nn.Module): The model that returns unnormalized logits.
        x (Variable): Input variable. (output of the softmax)
        logits (Variable): Output of the model (input to the softmax layer)
        eps (float): Attack step size.
        xi (float): the finite difference parameter
    Returns:
        adv_x (Variable):
            Generated adversarial example.
        lds (Variable):
            The LDS value.
    Reference:
        Distributional Smoothing With Virtual Adversarial Training
        https://arxiv.org/abs/1507.00677
    """
    y1 = Variable(pred.data)
    d = Variable(torch.randn(x.size()))
    for i in range(iterations):
        y2 = model(x + xi * d)
        if apply_softmax:
            y2 = F.softmax(y2, dim=1)
        kl = kl_div(y1, y2)
        gradient = grad(kl, x)[0]
        d = Variable(gradient.data)
        if batch_norm:
            d = l2_batch_norm(d)
    adv_x = x + eps * d
    return adv_x



from __future__ import absolute_import, division, print_function

import numpy as np
import torch
from torch.autograd import Variable
from detox.attack import Attack
from detox.attack.functional import fgm, fgsm


class FGM(Attack):
    """
    Fast Gradient Method. It extends the original implementation with
    additional options of L1, L2, and L-inf norms (sign).
    Reference:
        Explaining and Harnessing Adversarial Examples,
        Goodfellow et al., CoRR2014, https://arxiv.org/abs/1412.6572
    """
    def __init__(self, model, eps=0.3, ord=np.inf,
                 clip_min=None, clip_max=None, apply_softmax=False):
        super(FGM, self).__init__(model)
        if not isinstance(model, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.model = model
        self.eps = eps
        self.ord = ord
        self.clip_min = clip_min
        self.clip_max = clip_max
        self.targeted = False
        self.apply_softmax = apply_softmax

    def generate(self, x, y=None, target=None):
        pred = self.model(x)
        if self.apply_softmax:
            pred = F.softmax(pred, dim=1)

        if target is not None:
            y = target
            self.targeted=True
        elif y is not None:
            y = y
            self.targeted = False
        else:
            _, y = torch.max(pred, 1)
            y = Variable(y.data)

        return fgm(x, pred, y=y, eps=self.eps, ord=self.ord,
                   clip_min=self.clip_min, clip_max=self.clip_max,
                   targeted=self.targeted)


class FGSM(Attack):
    """
    Fast Gradient Sign Method.
    Reference:
        Explaining and Harnessing Adversarial Examples,
        Goodfellow et al., CoRR2014, https://arxiv.org/abs/1412.6572
    """
    def __init__(self, model, eps=0.3, ord=np.inf,
                 clip_min=None, clip_max=None, apply_softmax=False):
        super(FGSM, self).__init__()
        self.model = model
        self.eps = eps
        self.ord = ord
        self.clip_min = clip_min
        self.clip_max = clip_max
        self.apply_softmax = apply_softmax

    def generate(self, x):
        pred = self.model(x)
        if self.apply_softmax:
            pred = F.softmax(pred, dim=1)

        return fgsm(x, pred, eps=self.eps, ord=self.ord,
                    clip_min=self.clip_min, clip_max=self.clip_max)



from __future__ import absolute_import, division, print_function

from detox.attack import Attack
from detox.attack.functional import saliency_map, jacobian
import numpy as np

class JSMA(Attack):
    r"""
    Jacobian-based Saliency Map Approach (JSMA). This method crafts adversarial
    samples based on an understanding of the mapping between inputs and outputs
    of DNNs.
    Reference:
        The Limitations of Deep Learning in Adversarial Settings,
        Papernot et al., https://arxiv.org/abs/1511.07528
    """
    def __init__(self, model, theta=1., gamma=0.1, clip_min=None, clip_max=None,
                 apply_softmax=False):
        super(JSMA, self).__init__()
        if not isinstance(model, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.model = model
        self.theta = theta
        self.gamma = gamma
        self.targeted = False
        self.clip_min = clip_min
        self.clip_max = clip_max
        self.apply_softmax = apply_softmax

    def generate(self, x, y=None, target=None):

        if target is not None:
            y = target
            self.targeted = True
        elif y is not None:
            y = y
            self.targeted = False
        else:
            # optain label with the first prediction.
            prob = self.model(x)
            if self.apply_softmax:
                prob = F.softmax(prob, dim=1)
            _, y = torch.max(prob.data, 1)

        adv_x_batch = Variable(torch.zeros_like(x.data))
        for i in range(x.size(0)):
            adv_x = jsma(self.model, x[i], y[i], gamma=self.gamma,
                         apply_softmax=self.apply_softmax)
            adv_x_batch[i] = adv_x

        return adv_x_batch



from __future__ import absolute_import, division, print_function

import math
import numpy as np
import torch
import torch.nn.functional as F


def l2_batch_norm(x, eps=1e-12):
    shape = x.size()
    x = x.view(shape[0], -1)
    x /= eps + torch.max(torch.abs(x), 1, keepdim=True)[0]
    square = torch.sum(torch.pow(x, 2), 0, keepdim=True)
    x_inv_norm = x / torch.rsqrt(math.sqrt(eps) + square)
    x_norm = x * x_inv_norm
    return x_norm.view(shape)


def kl_div(p, q):
    kl = torch.sum(p * (torch.log(p + 1e-12) - torch.log(q + 1e-12)), 1)
    return kl


def clip_eta(eta, ord, eps):
    if ord not in (np.inf, 1, 2):
        raise ValueError('ord must be one of np.inf, 1, 2')

    if ord == np.inf:
        norm = torch.clamp(eta, -eps, eps)
    else:
        norm = F.normalize(eta, ord)
    return eta * torch.clamp(eps / (norm + 1e-12), max=1.)


def tanh_rescale(x, min=-1., max=1.):
    return (torch.tanh(x) + 1) * 0.5 * (max - min) + min



from __future__ import absolute_import, division, print_function

import torch
from torch.autograd import Variable
from detox.attack import Attack
import numpy as np
import torch.nn.functional as F
from detox.attack.functional import fgm


class BasicIterative(Attack):
    """
    Basic Iterative Method (Kurakin et al. 2016).
    Reference:
        Adversarial Examples In The Phisical World,
        Kurakin et al., https://arxiv.org/abs/1607.02533
    """
    def __init__(self, model, eps=0.3, ord=np.inf, n_iter=1,
                 clip_min=None, clip_max=None, apply_softmax=False):
        super(BasicIterative, self).__init__()
        if not isinstance(model, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.model = model
        self.n_iter = n_iter
        self.eps = eps
        self.ord = ord
        self.clip_min = clip_min
        self.clip_max = clip_max
        self.targeted = False
        self.apply_softmax = apply_softmax

    def generate(self, x, y=None, target=None):

        if target is not None:
            y = target
            self.targeted = True
        elif y is not None:
            y = y
            self.targeted = False
        else:
            # optain label with the first prediction.
            pred = self.model(x)
            if self.apply_softmax:
                pred = F.softmax(pred, dim=1)
            _, y = torch.max(pred, 1)
            y = Variable(y.data)

        eta = Variable(torch.zeros_like(x.data))
        # let's cook the perturbation
        for i in range(self.n_iter):
            adv_x = Variable(x.data + eta.data, requires_grad=True)
            pred = self.model(adv_x)
            if self.apply_softmax:
                pred = F.softmax(pred, dim=1)

            eta = fgm(adv_x, pred, y=y, eps=self.eps, ord=self.ord,
                      clip_min=None, clip_max=None, targeted=self.targeted)
            eta = eta - x

            if self.ord == np.inf:
                norm = torch.clamp(eta, -self.eps, self.eps)
            else:
                norm = F.normalize(eta, self.ord)
            eta = eta * self.eps / (norm + 1e-12)
        # add perturbation
        adv_x = x + eta

        # clip if necessary
        if (self.clip_min is not None) and (self.clip_max is not None):
            adv_x = torch.clamp(adv_x, self.clip_min, self.clip_max)
        return adv_x



from __future__ import absolute_import, division, print_function

import numpy as np
import torch
import torch.nn as nn
from torch.autograd import Variable
from detox.shim import is_volatile

# Tensor Related Functions


def to_gpu(x, *args, **kwargs):
    r""" Moves torch tensor to gpu if CUDA is available.
    Args:
        x (torch.Tensor): tensor to be moved to the GPU.
    Returns:
        torch.Tensor: Tensor moved to the GPU if CUDA is available otherwise the
                      original tensor.
    """
    return x.cuda(*args, **kwargs) if torch.cuda.is_available() else x


def get_layer(model, idx):
    """ Returns the layer of the corresponding index of the model."""
    return list(model.children())[idx]


def cut_model(model, idx):
    """ Makes a new model with the layers extracted from the input model. """
    nets = list(model.children())[:-idx]
    return TruncatedModel(nets)


class TruncatedModel(nn.Module):
    def __init__(self, parts):
        """parts is a list of nn.Module components"""
        super(TruncatedModel, self).__init__()
        self.features = nn.Sequential(*parts)

    def forward(self, x):
        x = self.features(x)
        return x



import torch.nn as nn
import torch.utils.model_zoo as model_zoo


__all__ = ['AE', 'ae']


model_urls = {
    'ae': 'TODO',
}


class Autoencoder(nn.Module):
    """ Autoencoder used for adversarial transformation networks attack.

    """
    def __init__(self, encoder=None, decoder=None):
        super(Autoencoder, self).__init__()
        if encoder is None:
            self.encoder = nn.Sequential(
                nn.Conv2d(1, 16, 3, stride=3, padding=1),  # b,16,10,10
                nn.ReLU(True),
                nn.MaxPool2d(2, stride=2),  # b,16,5,5
                nn.Conv2d(16, 8, 3, stride=2, padding=1),  # b,8,3,3
                nn.ReLU(True),
                nn.MaxPool2d(2, stride=1)  # b,8,2,2
            )
        if decoder is None:
            self.decoder = nn.Sequential(
                nn.ConvTranspose2d(8, 16, 3, stride=2),  # b,16,5,5
                nn.ReLU(True),
                nn.ConvTranspose2d(16, 8, 5, stride=3, padding=1),  # b,8,15,15
                nn.ReLU(True),
                nn.ConvTranspose2d(8, 1, 2, stride=2, padding=1),  # b,1,28,28
                nn.Tanh()
            )

    def forward(self, x):
        x = self.encoder(x)
        x = self.decoder(x)
        return x


def autoencoder(pretrained=False, **kwargs):
    r"""Adversarial Transformation Networks

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = AE(**kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['ae']))
    return model



from __future__ import absolute_import, division, print_function

from detox.nn.autoencoder import Autoencoder, autoencoder
from detox.nn.wgan import WGAN, wgan

__all__ = [
    # Classes
    'Autoencoder',
    'WGAN',
    # Functionals
    'autoencoder',
    'wgan',
]



from __future__ import absolute_import, division, print_function

import numpy as np
from detox.attack import Attack
from detox.attack import DeepFool


class UAP(Attack):
    """
    Universal adversarial perturbations
    Reference:
        Universal adversarial perturbations,
        Moosavi-Dezfooli et al., https://arxiv.org/abs/1610.08401
    """
    def __init__(self, model, delta=0.2, max_iter=None, xi=10, ord=np.inf,
                 n_candidates=10, overshoot=0.02, max_iter_df=10):
        super(UAP, self).__init__()
        if not isinstance(model, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.model = model
        self.delta = delta
        self.n_candidates = n_candidates
        self.overshoot = overshoot
        self.max_iter = max_iter
        self.clip_min = clip_min
        self.clip_max = clip_max
        self.attack = DeepFool(model, n_candidates, overshoot,
                               max_iter=max_iter_df)

    def proj_lp(self, v, xi, p):
        # Project on the lp ball centered at 0 and of radius xi
        # SUPPORTS only p = 2 and p = Inf for now
        if p == 2:
            v = v * min(1, xi / np.linalg.norm(v.flatten(1)))
            # v = v / np.linalg.norm(v.flatten(1)) * xi
        elif p == np.inf:
            v = np.sign(v) * np.minimum(abs(v), xi)
        else:
             raise ValueError('Values of p different from 2 and Inf are currently not supported...')
        return v

    def generate(self, x):

        v = 0
        fooling_rate = 0.0
        n_images = x.size(0)

        i = 0
        while fooling_rate < 1 - delta and i < self.max_iter:
            # ompute the perturbation increments sequentially
            for k in range(n_images):
                cur_img = x[k:(k+1), :, :, :]
                f = self.model(x)
                _, y = torch.max(f, 1)
                f_adv = self.model(x + v)
                _, y_adv = torch.max(f_adv, 1)

                if int(y) == int(y_adv):
                    # Compute adversarial perturbation
                    adv_x, perturbation = self.deepfool(x + v)
                    v = v + perturbation

                    # Project on l_p ball
                    v = self.project_lp(v, xi, p)

            # add perturbation to x
            adv_x = x + v

            pred_orig = torch.zeros([n_images])
            pred_pert = torch.zeros([n_images])

            batch_size = 100.0
            n_batches = int(np.ceil(n_images / batch_size))

            # Compute the estimated labels in batches
            for b in range(n_batches):
                m = (b * batch_size)
                M = min((b + 1) * batch_size, n_images)
                f = self.model(x[m:M, ...])
                f_adv = self.model(adv_x[m:M, ...])

                _, y = torch.max(f, 1)
                _, y_adv = torch.max(f_adv, 1)
                pred_orig[m:M] = y
                pred_pert[m:M] = y_adv

            # Compute the fooling rate
            fooling_rate = torch.sum(pred_pert != pred_orig) / n_images
            i = i + 1

        return v



from __future__ import absolute_import, division, print_function

import logging

name = "dtox"
__version__ = '0.1.0'

# Default logger to prevent 'No handler found' warning.
logging.getLogger(__name__).addHandler(logging.NullHandler())

from detox.functional import *
from detox.metric import *

def set_seed(seed):
    """
    Sets seeds of torch, numpy, and torch.cuda (if available).
    :param int rng_seed: The seed value.
    """
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)



from __future__ import absolute_import, division, print_function

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
from detox.attack import Attack


class OnePixel(Attack):
    """
    One-Pixel Attack is a semi-blackbox Attack which only requires the
    probability output of the model. Rather than having a constraint on the
    intensity of the perturbation, it considers the number of pixels as the
    limitation.
    Reference:
        One pixel attack for fooling deep neural networks,
        Su et al., https://arxiv.org/abs/1710.08864
    """
    def __init__(self, model, F=0.5, C=0.5, n_iter=75, apply_softmax=False):
        super(OnePixel, self).__init__()
        if not isinstance(model, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.model = model
        self.targeted
        self.F = F
        self.C = C
        self.n_iter = n_iter
        self.apply_softmax = apply_softmax

    def generate(self, x, y=None, target=None):
        """
        iteration of mutation, crossover and selection
        """
        pred = self.model(x)
        if self.apply_softmax:
            pred = F.softmax(pred, dim=1)

        if target is not None:
            y = target
            self.targeted = True
        elif y is not None:
            y = y
            self.targeted = False
        else:
            _, y = torch.max(pred, 1)
            y = Variable(y.data)

        batch_size, C, H, W = x.size()
        flatten = x.view(batch_size, -1)
        #
        # for b in range(batch_size):
        #     flat_img  = x[b]
        #     fitness = list(map(objective, flat_img))
        #
        #     for i in range(max_iter):
        #         mutants = mutate(pop, fitness, F)
        #         trial = crossover(pop, mutants, C)
        #         pop, fitness = select(objective, pop, fitness, trial)
        #         print(pop)

    def objective(self):
        return 1

    def random_unique(self, length, count, include=[], exclude=[]):
        indices = list(include)
        while len(indices) < count:
            next = np.random.randint(length)
            while next in indices or next in exclude:
                next = np.random.randint(length)
            indices.append(next)
        return indices

    def mutate(self, pop, fitness, F):
        pop = np.array(pop, copy=False)
        fitness = np.array(fitness, copy=False)
        pop_size = len(pop)
        de_idx = lambda count: np.array(
            [random_unique(pop_size, count, exclude=[i])
             for i in range(pop_size)]
        )

        # Could be more methods... but
        parents = pop[de_idx(3)]
        return parents[:,0] + F * (parents[:,1] - parents[:,2])

    def crossover(self, pop, mutant, C):
        # mutant enters the trial victor with probability C
        pop = np.array(pop, copy=False)
        mask = np.random.random_sample(mutant.shape) < C
        return np.where(mask, mutant, pop)

    def select(self, objective, pop, fitness, trial, stochastic=False):
        """
            objective: function
            pop: original population
            fitness: original population fitness
            trial_pop: trial population
        """
        pop = np.array(pop, copy=False)
        fitness = np.array(fitness, copy=False)

        # objective function is evaluated for each trial vector
        trial_fitness = np.array(list(map(objective, trial)))

        # storn-price selection method
        mask = fitness < trial_fitness
        new_fitness = np.where(mask, fitness, trial_fitness)
        new_pop = np.where(mask, pop, trial)
        if stochastic:
            new_fitness = np.array(list(map(objective, new_pop)))
        return new_pop, new_fitness



from __future__ import absolute_import, division, print_function

from detox.attack import Attack


# TODO
class LBFGS(Attack):
    """ Formalizes the generation of adversarial examples as an optimization
    problem.
    Reference:
        Exploring the Space of Adversarial Images,
        Tabacof et al., https://arxiv.org/abs/1510.05328
    """
    def __init__(self, model, lr=0.01, reg=0.01, momentum=0.9, max_iter=10):
        super(LBFGS, self).__init__()
        if not isinstance(model, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.model = model
        self.lr = lr
        self.reg = reg
        self.momentum = momentum
        self.max_iter = max_iter

    def criterion(self, x, probs, original_x, targets):
        model_loss = F.cross_entropy(probs, targets)
        image_loss = F.mse_loss(x, original_x)
        loss = model_loss + self.reg * image_loss
        return loss

    def generate(self, x, y=None, target=None):

        def iter_x(self, x): # TODO replace this with adv_x
            yield x

        adv_x = Varialbe(x.data, requires_grad=True)
        optimizer = SGD(iter_x(x), lr=self.lr, momentum=self.momentum)

        # TODO: clamp images
        probs = self.model(x)
        if apply_softmax:
            probs = F.softmax(probs, dim=1)
        preds = torch.max(probs.data, 1)[1]
        original_preds = preds.clone()

        # TODO: non-target
        targets = Variable(torch.LongTensor([target] * self.batch_size))

        i = 0
        while i < self.max_iter and torch.all(original_preds != preds):
            optimizer.zero_grad()

            # TODO: clamp loss so that all pixels are in range [0, 1]
            probs = self.model(adv_x)
            if self.apply_softmax:
                probs = F.softmax(probs, dim=1)
            loss = F.cross_entropy(adv_x, probs, x, targets)
            pred_loss, preds = torch.max(probs.data, 1)

            loss = self.criterion(adv_x, probs, x, targets)
            loss.backward()
            optimizer.step()

        return adv_x



import torch.nn as nn

class Generator(nn.Module):
    def __init__(self):
        super(Generator, self).__init__()

        preprocess = nn.Sequential(
            nn.Linear(128, 4*4*4*DIM),
            nn.ReLU(True),
        )
        block1 = nn.Sequential(
            nn.ConvTranspose2d(4*DIM, 2*DIM, 5),
            nn.ReLU(True),
        )
        block2 = nn.Sequential(
            nn.ConvTranspose2d(2*DIM, DIM, 5),
            nn.ReLU(True),
        )
        deconv_out = nn.ConvTranspose2d(DIM, 1, 8, stride=2)

        self.block1 = block1
        self.block2 = block2
        self.deconv_out = deconv_out
        self.preprocess = preprocess
        self.sigmoid = nn.Sigmoid()

    def forward(self, input):
        output = self.preprocess(input)
        output = output.view(-1, 4*DIM, 4, 4)
        output = self.block1(output)
        output = output[:, :, :7, :7]
        output = self.block2(output)
        output = self.deconv_out(output)
        output = self.sigmoid(output)
        return output.view(-1, OUTPUT_DIM)

class Discriminator(nn.Module):
    def __init__(self):
        super(Discriminator, self).__init__()

        main = nn.Sequential(
            nn.Conv2d(1, DIM, 5, stride=2, padding=2),
            nn.ReLU(True),
            nn.Conv2d(DIM, 2*DIM, 5, stride=2, padding=2),
            nn.ReLU(True),
            nn.Conv2d(2*DIM, 4*DIM, 5, stride=2, padding=2),
            nn.ReLU(True),
        )
        self.main = main
        self.output = nn.Linear(4*4*4*DIM, 1)

    def forward(self, input):
        input = input.view(-1, 1, 28, 28)
        out = self.main(input)
        out = out.view(-1, 4*4*4*DIM)
        out = self.output(out)
        return out.view(-1)


class ResBlock(nn.Module):
    def __init__(self, dim):
        super(ResBlock, self).__init__()
        self.res_block = nn.Sequential(
            nn.ReLU(True),
            nn.Conv1d(dim, dim, 5, padding=2),  # nn.Linear(DIM, DIM),
            nn.ReLU(True),
            nn.Conv1d(dim, dim, 5, padding=2),  # nn.Linear(DIM, DIM),
        )

    def forward(self, input):
        output = self.res_block(input)
        return input + (0.3 * output)


class Generator(nn.Module):
    def __init__(self, dim, seq_len, vocab_size):
        super(Generator, self).__init__()
        self.dim = dim
        self.seq_len = seq_len

        self.fc1 = nn.Linear(128, dim * seq_len)
        self.block = nn.Sequential(
            ResBlock(dim),
            ResBlock(dim),
            ResBlock(dim),
            ResBlock(dim),
            ResBlock(dim),
        )
        self.conv1 = nn.Conv1d(dim, vocab_size, 1)
        self.softmax = nn.Softmax()

    def forward(self, noise):
        batch_size = noise.size(0)
        output = self.fc1(noise)
        # (BATCH_SIZE, DIM, SEQ_LEN)
        output = output.view(-1, self.dim, self.seq_len)
        output = self.block(output)
        output = self.conv1(output)
        output = output.transpose(1, 2)
        shape = output.size()
        output = output.contiguous()
        output = output.view(batch_size * self.seq_len, -1)
        output = self.softmax(output)
        # (BATCH_SIZE, SEQ_LEN, len(charmap))
        return output.view(shape)


class Discriminator(nn.Module):
    def __init__(self, dim, seq_len, vocab_size):
        super(Discriminator, self).__init__()
        self.dim = dim
        self.seq_len = seq_len

        self.block = nn.Sequential(
            ResBlock(dim),
            ResBlock(dim),
            ResBlock(dim),
            ResBlock(dim),
            ResBlock(dim),
        )
        self.conv1d = nn.Conv1d(vocab_size, dim, 1)
        self.linear = nn.Linear(seq_len * dim, 1)

    def forward(self, input):
        # (BATCH_SIZE, VOCAB_SIZE, SEQ_LEN)
        output = input.transpose(1, 2)
        output = self.conv1d(output)
        output = self.block(output)
        output = output.view(-1, self.seq_len * self.dim)
        output = self.linear(output)
        return output



from __future__ import absolute_import, division, print_function

import torch
from torch import optim
from torch.autograd import Variable
import numpy as np
from detox.attack import Attack
from detox.util import tanh_rescale, torch_arctanh, l2_dist


class CarliniWagnerL2(object):
    def __init__(self, confidence=0, lr=1e-3, init_const=1,
                 binary_search_steps=5, max_iter=1000, abort_early=True,
                 clip_min=-1, clip_max=1):
        self.confidence = confidence
        self.init_const = init_const
        self.lr = lr
        self.binary_search_steps = binary_search_steps
        self.repeat = binary_search_steps >= 10
        self.max_iter = max_iter
        self.abort_early = abort_early
        self.clip_min = clip_min
        self.clip_max = clip_max

    def _compare(self, output, target):
        if not isinstance(output, (float, int, np.int64)):
            output = np.copy(output)
            if self.targeted:
                output[target] -= self.confidence
            else:
                output[target] += self.confidence
            output = np.argmax(output)
        if self.targeted:
            return output == target
        else:
            return output != target

    def _criterion(self, output, target, dist, scale_const):
        # compute the probability of the label class versus the maximum other
        real = (target * output).sum(1)
        other = ((1. - target) * output - target * 10000.).max(1)[0]
        if self.targeted:
            # optimize for making the other class most likely
            loss1 = torch.clamp(other - real + self.confidence, min=0.)
        else:
            # optimize for making this class least likely.
            loss1 = torch.clamp(real - other + self.confidence, min=0.)
        loss1 = torch.sum(scale_const * loss1)
        loss2 = dist.sum()
        loss = loss1 + loss2
        return loss

    def _step(self, optimizer, x, eta, target, scale_const, orig):
        # create adv_x and convert to tanh space
        adv_x = x + eta
        adv_x = tanh_rescale(adv_x, self.clip_min, self.clip_max)

        output = self.model(adv_x)
        dist = l2_dist(adv_x, orig, keepdim=False)
        loss = self._criterion(output, target, dist, scale_const)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        return loss.data[0], dist, output, adv_x

    def generate(self, input, target, batch_idx=0):
        batch_size = input.size(0)

        # set the lower and upper bounds accordingly
        lower_bound = np.zeros(batch_size)
        scale_const = np.ones(batch_size) * self.initial_const
        upper_bound = np.ones(batch_size) * 1e10

        # placeholders for the best results
        o_best_l2 = [1e10] * batch_size
        o_best_score = [-1] * batch_size
        o_best_attack = torch.zeros_like(input)

        # convert to tanh space
        x = Variable(torch_arctanh(input), requires_grad=False)
        orig = tanh_rescale(x, self.clip_min, self.clip_max)

        # target variable TODO: efficient sparse computation
        target_onehot = torch.zeros(target.size() + (self.num_classes,))
        if self._cuda:
            target_onehot = target_onehot.cuda()
        target_onehot.scatter_(1, target.unsqueeze(1), 1.)
        target = Variable(target_onehot, requires_grad=False)

        # initialize noise
        eta = torch.zeros(x.size()).float()
        if self._cuda:
            eta = eta.cuda()
        eta = Variable(eta, requires_grad=True)

        optimizer = optim.Adam([eta], lr=self.lr)

        for search_step in range(self.binary_search_steps):
            best_l2 = [1e10] * batch_size
            best_score = [-1] * batch_size

            # The last iteration (if we run many steps) repeat the search once.
            if self.repeat and search_step == self.binary_search_steps - 1:
                scale_const = upper_bound

            scale_const_tensor = torch.from_numpy(scale_const).float()
            if self.cuda:
                scale_const_tensor = scale_const_tensor.cuda()
            scale_const_var = Variable(scale_const_tensor, requires_grad=False)

            prev_loss = 1e6
            for i in range(self.max_iter):
                # optimize the noise (eta)
                loss, dist, output, adv_x = self._step(
                    optimizer, x, eta, target, scale_const_var, orig)

                if self.abort_early and i % (self.max_steps // 10) == 0:
                    if loss > prev_loss * .9999:
                        print('Aborting early...')
                        break
                    prev_loss = loss

                # update the best result found so far TODO: recheck
                for b, (targ, out, l2) in enumerate(zip(target, output, dist)):
                    _, pred = torch.max(out)
                    if l2 < best_l2[b] and self._compare(pred, targ):
                        best_l2[b] = l2
                        best_score[b] = pred
                    if l2 < o_best_l2[b] and self._compare(pred, targ):
                        o_best_l2[b] = l2
                        o_best_score[b] = pred
                        o_best_attack[b] = adv_x[b]

            # adjust the constants
            for b in range(batch_size):
                if self._compare(best_score[b], target[b]) \
                   and best_score[b] != -1:
                    # success, divide const by two
                    upper_bound[b] = min(upper_bound[b], scale_const[b])
                    if upper_bound[b] < 1e9:
                        scale_const[b] = (lower_bound[b] + upper_bound[b]) / 2
                else:
                    # failure, either multiply by 10 if no solution found yet
                    #          or do binary search with the known upper bound
                    lower_bound[b] = max(lower_bound[b], scale_const[b])
                    if upper_bound[b] < 1e9:
                        scale_const[b] = (lower_bound[b] + upper_bound[b]) / 2
                    else:
                        scale_const[b] *= 10

        return o_best_attack



from __future__ import absolute_import, division, print_function

import sys

from setuptools import find_packages, setup

# Find package version.
for line in open('detox/__init__.py'):
    if line.startswith('__version__ = '):
        version = line.strip().split()[2][1:-1]

# Convert README.md to rst for display at pypi (TODO)
# When releasing on pypi, make sure pandoc is on your system:
# $ brew install pandoc          # OS X
# $ sudo apt-get install pandoc  # Ubuntu Linux
try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError, OSError) as e:
    sys.stderr.write('Failed to convert README.md to rst:\n  {}\n'.format(e))
    sys.stderr.flush()
    long_description = open('README.md').read()

# Remove badges since they will always be obsolete.
blacklist = ['Build Status',
             'LICENSE',
             'Latest Version',
             'travis-ci.org',
             'pypi.python.org']

long_description = '\n'.join(
    [line for line in long_description.split('\n')
     if not any(patt in line for patt in blacklist)])

setup(
    name='dtox',
    version=version,
    description='TODO',
    long_description=long_description,
    packages=find_packages(exclude=('tests*',)),
    url='https://github.com/KakaoBrain',
    author='KakaoBrain',
    author_email='kwk236@nyu.edu',
    install_requires=[
        'numpy>=1.7',
        'scipy>=0.19.0',
        'torch',
        'six>=1.10.0',
    ],
    extras_require={
        'notebooks': ['jupyter>=1.0.0'],
        'visualization': [
            'matplotlib>=1.3',
            'visdom>=0.1.4',
            'pillow',
        ],
        'test': [
            'pytest',
            'pytest-cov',
            'nbval',
            'visdom',
            'torchvision',
        ],
        'profile': ['prettytable'],
        'dev': [
            'torchvision',
            'flake8',
            'yapf',
            'isort',
            'pytest',
            'pytest-xdist',
            'nbval',
            'nbstripout',
            'pypandoc',
            'sphinx',
            'sphinx_rtd_theme',
        ],
    },
    tests_require=['flake8', 'pytest'],
    keywords='machine learning security adversarial attack'
             'defense deep learning pytorch',
    license='MIT License',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3.4',
    ],
    # yapf
)



from __future__ import absolute_import, division, print_function

import abc
import torch.nn as nn
import torch.utils.hooks as hooks
from collections import OrderedDict


class Attack(object):
    r"""Base class for all attack classes.
    Your models should also subclass this class.
    ``Attack`` allows pre and post hooks to provide a room for additional
    processes.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, **kwargs):
        self._pre_hooks = OrderedDict()
        self._hooks = OrderedDict()
        self.targeted = True
        self._cuda = False

    def __call__(self, *input, **kwargs):
        for hook in self._pre_hooks.values():
            hook(self, input)
        result = self.generate(*input, **kwargs)
        for hook in self._hooks.values():
            hook_result = hook(self, input, result)
            if hook_result is not None:
                raise RuntimeError("a post hook should always return None")
        return result

    def register_pre_hook(self, hook):
        r"""Registers a pre-hook on the module.
        The hook will be called every time before :func:`generate` is invoked.
        It should have the following signature::
            hook(module, input) -> None
        The hook should not modify the input.
        Returns:
            :class:`torch.utils.hooks.RemovableHandle`:
                a handle that can be used to remove the added hook by calling
                ``handle.remove()``
        """
        handle = hooks.RemovableHandle(self._pre_hooks)
        self._pre_hooks[handle.id] = hook
        return handle

    def register_hook(self, hook):
        r"""Registers a hook on the module.
        The hook will be called every time after :func:`generate` has
        computed an output.
        It should have the following signature::
            hook(module, input, output) -> None
        The hook should not modify the input or output.
        Returns:
            :class:`torch.utils.hooks.RemovableHandle`:
                a handle that can be used to remove the added hook by calling
                ``handle.remove()``
        """
        handle = hooks.RemovableHandle(self._hooks)
        self._hooks[handle.id] = hook
        return handle

    @abc.abstractmethod
    def generate(self, *input):
        r"""Defines the generative computation performed at every call.
        Should be overriden by all subclasses.
        .. note::
            Although the recipe for adversarial example generation needs to be
            defined within this function, one should call the :class:`Attack`
            instance afterwards instead of this since the former takes care of
            running the registered hooks while the latter silently ignores
            them.
        """
        raise NotImplementedError

    def cuda(self):
        r""" Allow the use of CUDA in training. """
        self._cuda = True
        return self

    def __repr__(self):
        tmpstr = self.__class__.__name__ + '\n'
        return tmpstr



from __future__ import absolute_import, division, print_function

from detox.attack import Attack


class BoundaryAttack(Attack):
    """
    TODO
    Reference:
        Decision-Based Adversarial Attacks: Reliable Attacks Against
        Black-Box Machine Learning Models
        Brendel et al., https://arxiv.org/abs/1712.04248
    """
    def __init__(self, model):
        super(BoundaryAttack, self).__init__()
        if not isinstance(model, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.model = model

    def generate(self, x):
        pass



from __future__ import absolute_import, division, print_function

import torch as torch
from torch.nn import functional as F
from torch.autograd import Variable, grad
from torch.autograd.gradcheck import zero_gradients
from detox.attack import Attack


class DeepFool(Attack):
    """
    DeepFool computes adversarial samples based on iterative linearization
    of the classifier.
    Reference:
        DeepFool: a simple and accurate method to fool deep neural networks,
        Moosavi-Dezfooli et al., https://arxiv.org/pdf/1511.04599.pdf
    """
    def __init__(self, model, n_candidates=10, overshoot=0.02, max_iter=50,
                 clip_min=None, clip_max=None):
        """
        Args:
            model: models must output the logits (activation before softmax).
            k: number of k candidate classes to be considered in the attack.
               Unless specified, the number is chosen according to the
               prediction confidence.
            overshoot: termination criterion
            max_iter: maximum number of iterations
        """
        super(DeepFool, self).__init__()
        if not isinstance(model, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.model = model
        self.n_candidates = n_candidates
        self.overshoot = overshoot
        self.max_iter = max_iter
        self.clip_min = clip_min
        self.clip_max = clip_max

    def generate(self, x):
        """
        x: BxHxWxC
        Returns perturbed image and the perturbation.
        """
        batch_size, C, H, W = x.size()
        adv_x = Variable(x.data, requires_grad=True)

        # record original prediction of the model
        logits = self.model(adv_x)
        f, idx = logits.sort()
        idx = idx.data
        k_0 = idx[:, 0]

        k_i = k_0
        i = 0
        r_tot = Variable(torch.zeros_like(x.data))
        while k_0.eq(k_i).any() and i < self.max_iter:

            # get gradients wrt adv_x
            grad_0 = grad(f[:, k_0], adv_x, retain_graph=True)[0]

            for b in range(batch_size):
                pert = None

                for k in range(1, self.n_candidates):
                    zero_gradients(adv_x)
                    # get gradients wrt adv_x
                    grad_k = grad(f[b, idx[b, k]], adv_x, retain_graph=True)[0]

                    w_k = grad_k - grad_0
                    f_k = f[b, idx[b, k]] - f[b, idx[b, 0]]

                    pert_k = (
                        torch.abs(f_k) /
                        torch.sum((F.normalize(w_k.view(1, -1), 2)) + 1e-12)
                    )

                    # determine which w_k to use
                    if pert is None or (pert_k < pert).any():
                        pert = pert_k
                        w = w_k

                r_i = (pert + 1e-12) * w / F.normalize(w, 2)
                r_tot[b, ...] = r_tot[b, ...] + r_i

            # add perturbation
            adv_x = adv_x + (1 + self.overshoot) * r_tot
            adv_x = Variable(adv_x.data, requires_grad=True)

            f = self.model(adv_x)
            _, k_i = torch.max(f, 1)
            k_i = k_i.data
            i += 1

        adv_x = (1 + self.overshoot) * r_tot + x
        # clip if necessary
        if (self.clip_min is not None) and (self.clip_max is not None):
            adv_x = torch.clamp(adv_x, self.clip_min, self.clip_max)
        return adv_x, r_tot



from __future__ import absolute_import, division, print_function

import torch
from torch.autograd import Variable, grad
import numpy as np
import torch.nn.functional as F
from detox.attack import Attack
from detox.util import clip_eta


class MomentumIterative(Attack):
    """
    Momentum Iterative Method integrates the momentum term into the iterative
    process for attacks. Momentum stabilizes the update directions and escape
    from poor local maxima during the iterations, resulting in more
    transferable adversarial examples.
    Reference:
        Boosting Adversarial Attacks with Momentum,
        Dong et al., https://arxiv.org/pdf/1710.06081.pdf
    """
    def __init__(self, model, eps=0.3, ord=np.inf, n_iter=10, lr=0.05,
                 decay_factor=1.0, clip_min=None, clip_max=None,
                 apply_softmax=False):
        super(MomentumIterative, self).__init__()
        if not isinstance(model, nn.Module):
            raise ValueError('model must be an instance of nn.Module')
        self.model = model
        self.n_iter = n_iter
        self.eps = eps
        self.ord = ord
        self.lr = lr
        self.decay_factor = decay_factor
        self.clip_min = clip_min
        self.clip_max = clip_max
        self.apply_softmax = apply_softmax
        self.targeted = False

    def generate(self, x, y=None, target=None):

        if target is not None:
            y = target
            self.targeted = True
        elif y is not None:
            y = y
            self.targeted = False
        else:
            # optain label with the first prediction.
            pred = self.model(x)
            if self.apply_softmax:
                pred = F.softmax(pred, dim=1)
            _, y = torch.max(pred, 1)
            y = Variable(y.data)

        # initialize
        momentum = 0
        adv_x = Variable(x.data, requires_grad=True)

        # let's cook the perturbation
        for i in range(self.n_iter):
            pred = self.model(adv_x)
            if self.apply_softmax:
                pred = F.softmax(pred, dim=1)

            loss = F.cross_entropy(pred, y, size_average=False)
            if self.targeted:
                loss = -loss
            # get gradients wrt x
            gradient = grad(loss, adv_x)[0]

            # normalize gradient and add to accumulated gradient
            gradient /= (1e-12 + torch.mean(torch.abs(gradient)))

            momentum = self.decay_factor * momentum + gradient
            if self.ord == np.inf:  # L-inf (sign)
                norm_grad = torch.sign(momentum)
            else:  # Lp normalization
                norm_grad = F.normalize(momentum, self.ord)
            scaled_grad = self.lr * norm_grad
            adv_x = adv_x + scaled_grad
            adv_x = x + clip_eta(adv_x - x, self.ord, self.eps)

            # clip if necessary
            if (self.clip_min is not None) and (self.clip_max is not None):
                adv_x = torch.clamp(adv_x, self.clip_min, self.clip_max)
            adv_x = Variable(adv_x.data, requires_grad=True)  # stop gradient

        return adv_x
